/*
 *
 * This file is part of 3D Dice Box
 *
 * 3D Dice Box is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 3D Dice Box is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with 3D Dice Box.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*
 *
 * Copyright sbkaf 2025
 * All Rights Reserved.
 *
 */

import DiceBox from "./dice-box/dice-box.es.js";

const defaultDiceCode = "1d4 + 1d6 + 1d8 + 1d10 + 1d12 + 1d20";
const defaultThemeColor = "#21613e";
// some themes removed because of app size...
const diceToTheme = {
  d2: ["default-extras"],
  d4: [
    "default",
    /*
    "blueGreenMetal",
    "diceOfRolling",
    "gemstone",
    "gemstoneMarble",
    "rock",
    "rust",
    "smooth",
    "wooden",
    */
  ],
  d6: [
    "default",
    /*
    "blueGreenMetal",
    "diceOfRolling",
    "gemstone",
    "gemstoneMarble",
    "rock",
    "rust",
    "smooth",
    "wooden",
    */
  ],
  d8: [
    "default",
    /*
    "blueGreenMetal",
    "diceOfRolling",
    "gemstone",
    "gemstoneMarble",
    "rock",
    "rust",
    "smooth",
    "wooden",
    */
  ],
  d10: [
    "default",
    /*
    "blueGreenMetal",
    "diceOfRolling",
    "gemstone",
    "gemstoneMarble",
    "rock",
    "rust",
    "smooth",
    "wooden",
    */
  ],
  d12: [
    "default",
    /*
    "blueGreenMetal",
    "gemstone",
    "gemstoneMarble",
    "rock",
    "rust",
    "smooth",
    "wooden",
    */
  ],
  d20: [
    "default",
    /*
    "blueGreenMetal",
    "diceOfRolling",
    "rock",
    "rust",
    "smooth",
    "wooden",
    */
  ],
  d100: [
    "default",
    /*
    "blueGreenMetal",
    "diceOfRolling",
    "gemstone",
    "gemstoneMarble",
    "rock",
    "rust",
    "smooth",
    "wooden",
    */
  ],
  fate: [
    "default-extras",
    /*
    "diceOfRolling-fate",
    */
  ],
  pip: [
    "default-extras",
    /*
    "smooth-pip",
    */
  ],
  boost: ["genesys"],
  setback: ["genesys"],
  ability: ["genesys"],
  difficulty: ["genesys"],
  challenge: ["genesys"],
  proficiency: ["genesys"],
};
const themes = getUniqueValues(diceToTheme);

let TYPE_counter = 1
const TYPE_DICE_CODE = TYPE_counter ++
const TYPE_DICE_ROLL = TYPE_counter ++

// global vars
let diceCode = defaultDiceCode
let diceCodeLastChange = 0
let rollComplete = true;
let diceBox
let log = []
const toastBootstrap = bootstrap
  .Toast
  .getOrCreateInstance(
    elem('liveToast')
  )

main()
function main() {
  setDieTypeOptions()
  init3dRolling()
  initDiceCodeTextarea()
  initNewDieButton()
  initWebxdc()
}

function initWebxdc() {
  window.webxdc.setUpdateListener((update) => {
    switch(update.payload.type) {
      case TYPE_DICE_CODE:
        if (diceCodeLastChange < update.payload.at) {
          diceCodeLastChange = update.payload.at
          diceCode = update.payload.diceCode
          elem('diceCode').value = diceCode
        }
        break
      case TYPE_DICE_ROLL:
        log.push(update.payload)
    }
    log.sort((a, b) => {
      return b.at - a.at
    })
    elem('log').innerHTML = log
      .map(l => {
        return `
<a href="#" class="list-group-item" aria-current="true">
  <div class="d-flex w-100 justify-content-between">
    <h5 class="mb-1">${l.name} rolled <b>${l.sum}</b></h5>
    <small>${getDateStr(l.at)}</small>
  </div>
  <small>${l.text}</small>
</a>`
      })
      .join('')
  })
}

function getDateStr(d) {
  return (new Date(d)).toLocaleString()
}

function initDiceCodeTextarea() {
  elem('diceCode').onchange = storeDiceCode
}

function storeDiceCode() {
  const code = elem('diceCode').value
  webxdcSendUpdate({
    payload: {
      type: TYPE_DICE_CODE,
      diceCode: code,
      at: Date.now(),
    },
    summary: code,
  })
}

function webxdcSendUpdate(update) {
  window.webxdc.sendUpdate(update, '')
}

function initNewDieButton() {
  elem('addDice').onclick = () => {
    const amount = elem('diceAmount').value
    const color = elem('diceColor').value || ''
    const type = elem('diceType').value
    if (isNaN(amount)) {
      return
    }
    const diceCodeElem = elem('diceCode')
    diceCodeElem.value = mergeDiceCodes(diceCodeElem.value, getDieCode({
      type,
      amount,
      color,
    }))
    storeDiceCode()
    scrollToTop()
  }
}

function scrollToTop() {
  window.scrollTo(0, 0)
}

function getDieCode({type, amount, color}) {
  if (color) {
    return `${amount}${type}${color}`
  }
  return `${amount}${type}`
}

function mergeDiceCodes(a, b) {
  if (!a) {
    return b
  }
  return `${a} + ${b}`
}

function elem(id) {
  return document.getElementById(id)
}

function init3dRolling() {
  let loadedThemesCounter = 0;
  let initPromise = null;
  diceBox = new DiceBox("#dice-box", {
    assetPath: "/dice-box/",
    preloadThemes: themes,
    onThemeLoaded: () => {
      loadedThemesCounter++;
      if (loadedThemesCounter == themes.length) {
        initPromise.then(onReady);
      }
    },
    onRollComplete: (r) => {
      // if we do not wait here there will be graphic errors
      rollComplete = true;
    },
  });
  initPromise = diceBox.init();
}

function onReady() {
  elem("diceCode").value = diceCode
  elem("dice-box").onclick = roll;
  window.addEventListener("devicemotion", (event) => {
    if (
      event.rotationRate.alpha > 256 ||
      event.rotationRate.beta > 256 ||
      event.rotationRate.gamma > 256
    ) {
      roll();
    }
  });
  elem('loadingText').style.display = 'none'
  showToast({
    title: 'tap',
    body: 'tap to roll',
  })
}

function roll() {
  if (!rollComplete) {
    return;
  }
  rollComplete = false;
  const diceCode = elem("diceCode").value;
  const diceCodeArray = diceCode
    .split(/\s\+\s/)
    .map((diceCode) => {
      return diceCodeToObj(diceCode);
    })
    .filter((x) => !!x);
  diceBox.roll(diceCodeArray)
    .then(result => {
      let sum = 0
      result.forEach(r => {
        if (!isNaN(r.value)) {
          sum += r.value
        }
      })
      const text = result
        .map(r => {
          return `${r.dieType} to ${r.value}`
        })
        .join(', ')
      showToast({
        title: `Sum: ${sum}`,
        body: text,
      })
      const payload = {
        type: TYPE_DICE_ROLL,
        at: Date.now(),
        sum,
        text,
        name: window.webxdc.selfName,
      }
      webxdcSendUpdate({
        payload,
        info: `${payload.name} rolled ${payload.sum}`,
      })
    })
}

function showToast({title, body}) {
  elem('toast-title').innerText = title
  elem('toast-message').innerText = body
  toastBootstrap.show()
}

function setDieTypeOptions() {
  elem("diceType").innerHTML =
    getDiceNames().map(t => {
      return `<option>${t}</option>`
    })
    .join('\n')
}

function getDiceNames() {
  return Object.keys(diceToTheme);
}

function diceCodeToObj(diceCode) {
  const match = diceCode.match(/(\d+)([a-z][0-9a-z]+)(#[0-9a-fA-F]{6}|)/);
  if (!match) {
    alert(
      '"' +
        diceCode +
        '": wrong format\n\nuse: AmountDicename\nor\nAmountDicename#Color\n\ndice names:\n' +
        getDiceNames().join("\n"),
    );
    return match;
  }
  const theme = getThemeByDice(match[2]);
  if (!theme) {
    alert('"' + match[2] + '"-dice not found');
    return null;
  }
  return {
    qty: match[1],
    sides: match[2],
    theme,
    themeColor: match[3] || defaultThemeColor,
  };
}

function getThemeByDice(dice) {
  try {
    return diceToTheme[dice][0];
  } catch (e) {
    return null;
  }
}

function getUniqueValues(obj) {
  const uniqueValues = new Set();
  for (const key in obj) {
    if (Array.isArray(obj[key])) {
      obj[key].forEach((value) => uniqueValues.add(value));
    }
  }
  return [...uniqueValues];
}
